/********************************************************************\
 * readfile.c -- DWI glade-to-database XML file                     *
 * Copyright (C) 2002,2004 Linas Vepstas <linas@linas.org>          *
 * http://dwi.sourceforge.net                                       *
 *                                                                  *
 * This library is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU Lesser General Public       *
 * License as published by the Free Software Foundation; either     *
 * version 2.1 of the License, or (at your option) any later version.
 *                                                                  *
 * This library is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU Lesser General Public License for more details.              *
 *                                                                  *
 * You should have received a copy of the GNU Lesser General Public *
 * License along with this program; if not, contact:                *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
\********************************************************************/
/*
 * FUNCTION:
 * Parse an XML file that specifies the DWI glade-to-database coupling
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */
#include "config.h"

#include <string.h>
#include <stdlib.h>

#include <glib.h>
#include <glib/gfileutils.h>

#include <libxml/parser.h>

#include "perr.h"
#include "dui-initdb.h"

#include "action.h"
#include "database.h"
#include "duifield.h"
#include "duifield-gobj.h"
#include "duifield-hash.h"
#include "duifield-sql.h"
#include "duifieldmap.h"
#include "duifilter.h"
#include "readfile.h"
#include "report.h"
#include "parse-utils.h"

#define SUPPORT_OBSOLETE_MARKUP

struct DuiParser_s 
{
	/** XML file parser */
	xmlSAXHandler sax;

	/** Top-level interface anchor */
	DuiInterface *dui;

	/** List of "user-defined" element handler plugins for the parser */
	GList *element_plugins;
	
	/** List of plugins that can handle specific types of fields. */
	GList *field_plugins;
	
	/** List of plugins that get called when a new row/table is declared.a
	 * These are typically used to handle sql-like "where" terms. */
	GList *table_plugins;
	
	/** List of notifiers for the creation of stuff ... 
	 * XXX should be replaced by a 'real' notifier framework.
	 */
	GList *struct_plugins;

	/** Current database connection that is being used */
	DuiDatabase *database;
	
	/** Current action that is being parsed */
	DuiAction  *action;

	/** If in a row_match context, the matcher the row widget */
	DuiTxnQuery *query;

	/** Current report that is being parsed */
	DuiReport *report;

	/** If in a row context, the name of the row widget */
   DuiTxnReport *row;
	char * row_widget;

	/** Current filter definition */
	DuiFilter *filter;

	/** Current pipe definition */
	DuiFieldMap *fieldmap;
};


/* =================================================================== */
/* a really cheap and dirty thing .. */

static const char * 
de_escape (const char *val)
{
	if (!val) return NULL;

	if (!strcmp (val, "&gt;")) return ">";
	if (!strcmp (val, "&lt;")) return "<";

	return val;
}

/* =================================================================== */

static gboolean 
generic_field_handler (DuiParser *par, const char *typo, DuiField *fld, 
                       const char * row_wid, const char ** attrs)
{
	/* Go over the list of field plugins */
	GList *node;
	for (node=par->field_plugins; node; node=node->next)
	{
		DuiParserFieldPlugin *plg = node->data;
		if (!strcmp (typo, plg->fieldname))
		{
			gboolean rc;
			rc = plg->handler (plg->user_data, fld, row_wid, attrs);
			if (rc) return rc;
		}
	}
	return FALSE;
}

/* =================================================================== */

static void 
set_fm_source (DuiParser *par, DuiField *fld,
                const char ** attrs,
                const char * widname,
                const char *datakey,
                const char *src_key,
                const char *value)
{
	if (widname && !datakey)
	{
		generic_field_handler (par, "widget", fld, NULL, attrs);
	}
	else if (widname && datakey)
	{
		generic_field_handler (par, "widget_data", fld, NULL, attrs);
	}
	else if (src_key)
	{
		dui_field_set_hash_key (fld, src_key);
	}
	else
	{
		dui_field_set_const (fld, value);
	}
}

/* =================================================================== */

static void 
action_add_sql_term (DuiParser *par, 
                     const char ** attrs,
                           const char * widname,
                           const char *datakey,
                           const char *src_key,
                           const char *value,
                           const char *filter_name,
                           const char *field_name,
                           const char *compare_op)
{
	DuiFieldMap *fm = dui_field_map_new ();

	set_fm_source (par, &fm->source, attrs, widname, datakey, src_key, value);

	fm->filtername = g_strdup (filter_name);
	fm->filter = NULL;

	if (compare_op)
	{
		dui_field_set_where (&fm->target, field_name, compare_op);
	}
	else
	{
		dui_field_set_sql (&fm->target, field_name);
	}

	dui_action_add_term (par->action, fm);
}

/* =================================================================== */
/* Above, the action_add_*_term are set up so that the widget 
 * is the source of the data, and the target is somwere else,
 * e.g. the sqldb.  Below, the report_add_*_term are kind of 
 * the inverse: the widget is the target, and the sql is the
 * source. That's why these routines look painfully similar but
 * are not the same.
 */

static void
report_add_column (DuiParser *par, const char ** attrs,
                        const char * widname, 
                        const char * datakey, const char * arg,
                        const char * field, const char * key, 
                        const char * value, const char * filter)
{
	DuiFieldMap *col;

	if (!par) return;
	ENTER ("(wid=%s, field=%s, key=%s, val=%s)",
		widname, field, key, value);

	col = dui_field_map_new ();

	/* Setup the data source; where data will come from. */
	col->source.type = DUI_FIELD_NONE;
	if (field)
	{
		dui_field_set_sql (&col->source, field);
	}
	else if (key)
	{
		dui_field_set_hash_key (&col->source, key);
	}
	else if (value)
	{
		dui_field_set_const (&col->source, value);
	}

	col->filtername = g_strdup (filter);
	
	/* Set up the data target, where data will be copied to */
	col->target.type = DUI_FIELD_NONE;
	if (datakey)
	{
		generic_field_handler (par, "widget_data", &col->target, NULL, attrs);
	}
	else if (arg)
	{
		generic_field_handler (par, "widget_arg", &col->target, NULL, attrs);
	}
	else if (widname)
	{
		generic_field_handler (par, "widget", &col->target, NULL, attrs);
	}
	if (par->row)
	{
		dui_txnreport_add_term (par->row, col);
	}
	else
	{
		dui_report_add_term (par->report, col);
	}
   LEAVE (" ");
}

/* =================================================================== */
/* Set the row iterator widget.  This widget must
 * somehow be capable of reporting multiple rows
 * (ie. must have a way of creating a new row)
 * The "name" is used to identify this row as the target of an action.
 * The widgetname must be in the scope of the gladefile:root_widget tree
 * If nest is non-zero, then the iterator must be capable of displaying
 * a tree.  'nest' is how far down the tree we are.  match_col
 * and match-field are used to determine the parent node of the tree:
 * children are inserted under the parent whose 'match_col' contains
 * the value from the database 'match_field' or hash 'match_key' or
 * value 'match_value'.
 */


static DuiFieldMap *
add_row_matcher (const char * match_field,
                const char * match_key, 
                const char * match_value)
{
	DuiFieldMap *fm;

	ENTER ("(mfld=%s, mkey=%s, mval=%s)",
		match_field, match_key, match_value);

	/* Add the matching term as well */
	fm = dui_field_map_new();

	if (match_field)
	{
		dui_field_set_sql (&fm->source, match_field);
	}
	else if (match_key)
	{
		dui_field_set_hash_key (&fm->source, match_key);
	}
	else if (match_value)
	{
		dui_field_set_const (&fm->source, match_value);
	}

	fm->filtername = NULL;
	fm->filter = NULL;

   LEAVE ("()");
	return fm;
}

/* =================================================================== */

static void 
handle_report (DuiParser *par, const xmlChar **attrs)
{
	DuiReport *rpt;
	static const char * eltname = "<report>";
	const char * name = NULL;
	int i;

	// XXX A report needs to be nested in a window only if
	// if the report uses widgets.  We need to check for 
	// nesting integrity elsewhere.
	// CHECK_NEST (window, "<window>");

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("name",  name)
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	/* ------------------------------------------------------------------ */
	/* Set up the widget so that when its clicked or whatever, 
	 * we run our sql statement.
	 */
	rpt = dui_report_new (name);
	dui_interface_add_report (par->dui, rpt);  
	par->report = rpt;

	/* Notify interested parties about this report */
	GList *node;
	for (node=par->struct_plugins; node; node=node->next)
	{
		DuiParserStructPlugin *plg = node->data;
		if (plg->report_created)
		{
			(plg->report_created) (plg->user_data, rpt);
		}
	}
}

static void 
handle_report_end (DuiParser *par)
{
	par->report = NULL;
}

/* =================================================================== */

static void 
handle_row (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<row>";
	int i;
	const char * name = NULL;
	const char * nest = NULL;
	const char * match_field = NULL;
	const char * match_key = NULL;
	const char * match_value = NULL;
	const char * junk;
	int nest_level = 0;

	GENERIC_CHECKS ;
	CHECK_NEST2 (action, report, "<form>", "<report>");

	i = 0;
	while (attrs[i])
	{
		CASE ("name",         name)
		CASE ("nest",         nest)

		CASE ("match_field",  match_field)
		CASE ("match_key",    match_key)
		CASE ("match_value",  match_value)
		CASE ("widget",       junk)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* HMM, actually the valid forms are either match_field, key or
    * value, but only one of those.
    * If there's a widget, then there must(?) be a column. 
    * If there's an qof object, then there must(?) be a match-prop.
    */
	// REQUIRE2 (widgetname, qobj_type, "a widget or a QOF object type!");

	if (nest) nest_level = atoi (nest);
	
	DuiFieldMap *fm = add_row_matcher ( 
	                       match_field, match_key, match_value);

	/* Go over the list of fieldmap plugins */
	GList *node;
	for (node=par->table_plugins; node; node=node->next)
	{
		const char *rw;
		DuiParserTablePlugin *plg = node->data;
		rw = plg->handler (plg->user_data, fm, (const char **) attrs);
		if (rw)
		{
			par->row_widget = g_strdup (rw);
			break;
		}
	}

	/* XXX at this point, par->row_widget must be defined, else its 
	 * an error.  Typically, par->row_widget will be a widget or 
	 * a qobject type.
	 */
	REQUIRE (par->row_widget, "a widget or a QOF object type!");

	if (par->report)
	{
		par->row = dui_txnreport_new (name, nest_level);
		dui_report_add_row (par->report, par->row);
		dui_txnreport_add_match_term (par->row, fm);
	}

	if (par->action)
	{
		// Currently query is created when form is opened ... 
		// par->query = dui_txnquery_new();
		// dui_action_add_row(par->action, par->query);
		dui_txnquery_add_source_match_term (par->query, fm);
	}
}

static void 
handle_row_end (DuiParser *par)
{
	if (par->row_widget) g_free (par->row_widget);
	par->row_widget = NULL;
	par->row = NULL;
	// par->query = NULL;
}

/* =================================================================== */

static void 
handle_entry (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<entry>";
	int i;
	const char * name = NULL;
	const char * widgetname = NULL;
	const char * fieldname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * arg = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->report, "<report>");

	i = 0;
	while (attrs[i])
	{
		CASE ("name",   name)
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("arg",    arg)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		CASE ("field",  fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	// hack alert -- fixme -- 
	// a widget is not required if we are nested inside a row;
	// otherwise a widget is mandatory
	// REQUIRE (widgetname, "a widget!");
	// fix this so that we syntax-check for the one or the other.
	REQUIRE3 (fieldname, value, key, "a field or a key or a value!");

	if (!widgetname) widgetname = par->row_widget;
	report_add_column (par, (const char **) attrs, widgetname, datakey, arg,
	                       fieldname, key, value, filtername);
}


/* =================================================================== */
/* ********************* form handlers below ************************* */
/* =================================================================== */

/* =================================================================== */

static void 
handle_form (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<form>";
	int i;
	const char * formname = NULL;
	const char * reportname = NULL;
	const char * sql_db_name = NULL;
	const char * sql_querytype = NULL;
	const char * sql_tables = NULL;
	const char * key_to_sql_tables = NULL;
	DuiAction *act;
	DuiField *fld = NULL;

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("name",       formname)
			CASE ("type",       sql_querytype)
			CASE ("database",   sql_db_name)
			CASE ("tables",     sql_tables)
			CASE ("tables_key", key_to_sql_tables)
			CASE ("report",     reportname)
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	/* ------------------------------------------------------------------ */
	/* Set up the widget so that when its clicked or whatever, 
	 * we run our sql statement.
	 */
	act = dui_action_new (formname);
	dui_interface_add_action (par->dui, act);
	par->action = act;

	if (key_to_sql_tables)
	{
		fld = g_new0 (DuiField,1);
		fld->type = DUI_FIELD_NONE;
		dui_field_set_hash_key (fld, key_to_sql_tables);
	}
	else 
	if (sql_tables)
	{
		fld = g_new0 (DuiField,1);
		fld->type = DUI_FIELD_NONE;
		dui_field_set_const (fld, sql_tables);
	}
	if (fld || sql_querytype)
	{
		par->query = dui_txnquery_new();
		dui_action_add_row(par->action, par->query);
		dui_txnquery_set_table (par->query, fld, sql_querytype);
		dui_txnquery_set_database (par->query, par->database);
	}

	/* Action inherits the db & dbname from on-high */
	if (NULL == sql_db_name)
	{
		sql_db_name = dui_database_get_name (par->database);
	}
	if (sql_db_name && sql_db_name[0])
	{
		dui_action_set_database_name (act, sql_db_name);
	}
	dui_action_set_report (act, reportname);

	/* Notify interested parties about this action */
	GList *node;
	for (node=par->struct_plugins; node; node=node->next)
	{
		DuiParserStructPlugin *plg = node->data;
		if (plg->action_created)
		{
			(plg->action_created) (plg->user_data, act);
		}
	}
}

static void 
handle_form_end (DuiParser *par)
{
	if (!par->action) return;

	par->action = NULL;
	par->query = NULL;
}

/* =================================================================== */

static void 
handle_refresh (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<refresh>";
	int i;
	const char * reportname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("report", reportname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (reportname, "a report name!");
	dui_action_add_refresh (par->action, reportname);
}

/* =================================================================== */

static void 
handle_select (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<select>";
	int i;
	const char * fieldname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("field", fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (fieldname, "a data field!");
	{
		DuiFieldMap *fm;
		fm = dui_field_map_new();
		dui_field_set_const (&fm->target, fieldname);
		dui_action_add_term (par->action, fm);
	}
}

/* =================================================================== */

static void 
handle_insert (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<insert>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;
	const char * fieldname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		CASE ("field",  fieldname)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (fieldname, "a data field!");
	
	/* set up the action so that we fetch values from it */
	action_add_sql_term(par, (const char **) attrs, widgetname, 
	               datakey, key, value, filtername, fieldname, NULL);
}

/* =================================================================== */

static void 
handle_where (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<where>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * fieldname = NULL;
	const char * compareop = NULL;
	const char * key = NULL;
	const char * value = NULL;
	const char * filtername = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("key",    key)
		CASE ("value",  value)
		CASE ("filter", filtername)
		CASE ("field",  fieldname)
		CASE ("op",     compareop)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* Require either value or a key or a widget ... */
	REQUIRE3 (widgetname, key, value, "a widget or a key or a value!");
	REQUIRE (fieldname, "a data field!");

	/* Set up the action so that we fetch values from it */
	compareop = de_escape (compareop);
	if (!compareop) compareop = "=";
	action_add_sql_term(par, (const char **) attrs, widgetname,
		         datakey, key, value, filtername, fieldname, compareop);

}

/* =================================================================== */

static void 
handle_chain (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<chain>";
	int i;
	const char * actionname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("form", actionname)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require an action name ... */
	REQUIRE (actionname, "a form!");

	/* set up the action so that check values */
	dui_action_add_chain (par->action, actionname);
}

/* =================================================================== */

static void 
handle_pipe (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<pipe>";
	
	CHECK_NEST2 (action, report, "<form>", "<report>");

	DuiFieldMap *fm = dui_field_map_new ();
	par->fieldmap = fm;
	fm->source.type = DUI_FIELD_NONE;
	fm->target.type = DUI_FIELD_NONE;
	
	const char * filter_name = NULL;
	
	if (attrs)
	{
		int i=0;
		while (attrs[i])
		{
			CASE ("filter",     filter_name)
			CASE_UNKNOWN ;
			i += 2;
		}
	}

	fm->filtername = g_strdup (filter_name);
	fm->filter = NULL;
}

static void 
handle_pipe_end (DuiParser *par)
{
	if (par->action)
	{
		dui_action_add_term (par->action, par->fieldmap);
	} 
	else if (par->report)
	{
		if (par->row)
		{
			dui_txnreport_add_term (par->row, par->fieldmap);
		}
		else
		{
			dui_report_add_term (par->report, par->fieldmap);
		}
	}
	par->fieldmap = NULL;
}

/* =================================================================== */
/* parse generic terms, suitable for source or target fields */

static void
handle_generic_tgt_or_src (DuiParser *par, const char * eltname,
					            DuiField *fld, const xmlChar **attrs)
{
	int i;
	const char * typo = NULL;
	const char * column = NULL;
	const char * hashkey = NULL;
	const char * property = NULL;
	const char * object_instance_name = NULL;
	const char * junk;
	int colnum = -1;
	
	GENERIC_CHECKS ;
	CHECK_NEST (par->fieldmap, "<pipe>");
	
	i = 0;
	while (attrs[i])
	{
		CASE ("type",     typo)
		CASE ("column",   column)
		CASE ("key",      hashkey)
		CASE ("property", property)

		CASE ("object",   object_instance_name)

		/* Look for anything else, just to check syntax */
		CASE ("widget",   junk)
		CASE ("data",     junk)
		CASE ("arg",      junk)
		CASE ("value",    junk)
		CASE ("field",    junk)
		CASE ("op",       junk)
		CASE ("qof",      junk)   // XXXX
		CASE_UNKNOWN ;
		i += 2;
	}
	if (column) colnum = atoi (column);

	REQUIRE (typo, "a field type!");
	
	if (!strcmp (typo, "hash")) 
	{
		REQUIRE (hashkey,    "the hash key name!");
		dui_field_set_hash_key (fld, hashkey);
	}
	else
	if (!strcmp (typo, "gobject")) 
	{
		REQUIRE (object_instance_name, "an object!");
		REQUIRE (property, "a property!");
		dui_field_set_gobj (fld, object_instance_name, property);
	}

	/* Go over the list of field plugins */
	generic_field_handler (par, typo, fld, 
	                     par->row_widget, (const char **)attrs);
}

static void 
handle_src (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<src>";
	int i;
	const char * typo = NULL;
	const char * sqlfield = NULL;
	const char * const_value = NULL;
	
	DuiField *fld = &par->fieldmap->source;

	handle_generic_tgt_or_src (par, eltname, fld, attrs);
	
	i = 0;
	while (attrs[i])
	{
		CASE ("type",     typo)
		CASE ("field",    sqlfield)
		CASE ("value",    const_value)
		{}

		i += 2;
	}

	REQUIRE (typo, "a field type!");
	
	if (!strcmp (typo, "const")) 
	{
		REQUIRE (const_value,    "a value for the constant!");
		dui_field_set_const (fld, const_value);
	}
	else
	if (!strcmp (typo, "sql"))
	{
		REQUIRE (sqlfield,    "a SQL table field name!");
		dui_field_set_sql (fld, sqlfield);
	}

	if (!strcmp (fld->type, DUI_FIELD_NONE))
	{
		SYNTAX ("Unknown source field type=\"%s\"", typo);
	}
}

/* =================================================================== */

static void 
handle_tgt (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<tgt>";
	const char * typo = NULL;
	const char * sqlfield = NULL;
	const char * compareop = NULL;

	DuiField *fld = &par->fieldmap->target;

	handle_generic_tgt_or_src (par, eltname, fld, attrs);

	int i = 0;
	while (attrs[i])
	{
		CASE ("type",     typo)
		CASE ("field",    sqlfield)
		CASE ("op",       compareop)
		{}
		i += 2;
	}

	REQUIRE (typo, "a field type!");

	if (!strcmp (typo, "sql"))
	{
		REQUIRE (sqlfield,    "a SQL table field name!");

		if (compareop)
		{
			dui_field_set_where (fld, sqlfield, compareop);
		}
		else 
		{
			dui_field_set_sql (fld, sqlfield);
		}
		return;
	}

	if (!strcmp (fld->type, DUI_FIELD_NONE))
	{
		SYNTAX ("Unknown source field type=\"%s\"", typo);
	}
}

/* =================================================================== */

static void 
handle_database (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<database>";
	int i;
	const char * name = NULL;
	const char * provider = NULL;
	const char * dbname = NULL;
	const char * hostname = NULL;
	const char * username = NULL;
	const char * passwd = NULL;

	GENERIC_CHECKS ;

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("name",     name)
		CASE_STRDUP ("provider", provider)
		CASE_STRDUP ("dbname",   dbname)
		CASE_STRDUP ("host",     hostname)
		CASE_STRDUP ("username", username)
		CASE_STRDUP ("passwd",   passwd)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE  (provider, "a database provider!");
	REQUIRE2 (name, dbname, "an object instance name or a database name!");
	/* NULL dbnames, usernames, etc are allowed: these might be
	 * filled in via the object interface */

	par->database = dui_database_new (name, provider, dbname, hostname, username, passwd);
	dui_interface_add_database (par->dui, par->database);
}

/* =================================================================== */

static void 
handle_lookup (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<lookup>";
	const char * key = NULL; 
	const char * val = NULL; 
	int i;

	GENERIC_CHECKS ;
	CHECK_NEST (par->filter, "<filter>");

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("key",    key)
		CASE_STRDUP ("val",    val)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (key, "a lookup key!");
	REQUIRE (val, "a value that is looked up!");
	dui_filter_add_lookup (par->filter, key, val);
}

/* =================================================================== */

static void 
handle_filter (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<filter>";
	const char * filtername = NULL; 
	const char * reversed_name = NULL; 
	int i;

	GENERIC_CHECKS ;

	i = 0;
	while (attrs[i])
	{
		CASE_STRDUP ("name",    filtername)
		CASE_STRDUP ("revname", reversed_name)
		CASE_UNKNOWN ;
		i += 2;
	}

	REQUIRE (filtername, "a name for the new filter!");
	par->filter = dui_filter_new (filtername, reversed_name);
}

static void 
handle_filter_end (DuiParser *par)
{
	par->filter = NULL;
}

/* =================================================================== */

static void 
handle_dwi (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<dwi>";
	const char * app_name = NULL;
	int i;

	if (attrs)
	{
		i = 0;
		while (attrs[i])
		{
			CASE ("name", app_name)
			CASE_UNKNOWN ;
			i += 2;
		}

		dui_interfaced_set_name (par->dui, app_name);
		
	}
}

/* =================================================================== */
/* obsolete handlers */

#ifdef SUPPORT_OBSOLETE_MARKUP

/* =================================================================== */
/* obsolete handler */

static void 
handle_wtoobj (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<wtoobj>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * src_key = NULL;
	const char * value = NULL;
	const char * filter_name = NULL;
	const char * objname = NULL;
	const char * propname = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget", widgetname)
		CASE ("column", column)
		CASE ("data",   datakey)
		CASE ("key",    src_key)
		CASE ("value",  value)
		CASE ("filter", filter_name)
		
		CASE ("object",   objname)
		CASE ("property", propname)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or widget ... */
	REQUIRE3 (widgetname, src_key, value, "a widget or a key or a value!");
	REQUIRE (objname, "an object!");
	REQUIRE (propname, "a property!");

	/* set up the action so that we fetch values from it */
	DuiFieldMap *fm = dui_field_map_new ();

	set_fm_source (par, &fm->source, (const char **) attrs, 
	              widgetname, datakey, src_key, value);

	fm->filtername = g_strdup (filter_name);
	fm->filter = NULL;
	dui_field_set_gobj (&fm->target, objname, propname);

	dui_action_add_term (par->action, fm);
}

/* =================================================================== */
/* obsolete routine */

static void 
handle_wtokey (DuiParser *par, const xmlChar **attrs)
{
	static const char * eltname = "<wtokey>";
	int i;
	const char * widgetname = NULL;
	const char * column = NULL;
	const char * datakey = NULL;
	const char * src_key = NULL;
	const char * value = NULL;
	const char * filter_name = NULL;
	const char * dst_key = NULL;

	GENERIC_CHECKS ;
	CHECK_NEST (par->action, "<form>");

	i = 0;
	while (attrs[i])
	{
		CASE ("widget",  widgetname)
		CASE ("column",  column)
		CASE ("data",    datakey)
		// CASE ("key",  src_key)
		CASE ("value",   value)
		CASE ("filter",  filter_name)
		CASE ("dst_key", dst_key)
		CASE_UNKNOWN ;
		i += 2;
	}

	/* require either value or widget ... */
	REQUIRE2 (widgetname, value, "a widget or a value!");
	REQUIRE (dst_key, "a key!");

	DuiFieldMap *fm = dui_field_map_new ();

	set_fm_source (par, &fm->source, (const char **) attrs,
	               widgetname, datakey, src_key, value);

	fm->filtername = g_strdup (filter_name);
	fm->filter = NULL;

	dui_field_set_hash_key (&fm->target, dst_key);

	dui_action_add_term (par->action, fm);
}

#endif /* SUPPORT_OBSOLETE_MARKUP */

/* =================================================================== */

static void 
fatal_err (void *user_data, const char *msg, const char * m2)
{
	/* hack laert -- handle the elipses in the warning message */
	// DuiParser *par = (DuiInterface *) user_data;
	PERR ("DUI XML fatal parsing error: %s\n", m2);
}

static void 
err (void *user_data, const char *msg, const char *m2)
{
	// va_list ap;
	// DuiParser *par = (DuiInterface *) user_data;
	PERR ("DUI XML parsing error: %s\n", m2);

	// va_start (ap, msg);
}

static void 
warn (void *user_data, const char *msg, ...)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PERR ("DUI XML parse warning: %s\n", msg);
}

static void 
start_doc (void *user_data)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PINFO ("document start ptr=%p", user_data);
}

static void 
end_doc (void *user_data)
{
	// DuiParser *par = (DuiInterface *) user_data;
	PINFO ("document end");
}

/* =================================================================== */

#define DISPATCH(tok,handler)            \
   if (!strcmp (name, tok)) {            \
      handler (par, attrs);              \
   } else

#define TERMINATE(tok,handler)           \
   if (!strcmp (name, tok)) {            \
      handler (par);                     \
   } else

static void 
start_elt (void *user_data, const xmlChar *name, const xmlChar **attrs)
{
	DuiParser *par = (DuiParser *) user_data;
	gboolean found = TRUE;


	DISPATCH ("dwi",       handle_dwi)
	DISPATCH ("filter",    handle_filter)
	DISPATCH ("lookup",    handle_lookup)
	DISPATCH ("database",  handle_database)

	DISPATCH ("report",    handle_report)
	DISPATCH ("row",       handle_row)
	DISPATCH ("entry",     handle_entry)
	
	DISPATCH ("form",      handle_form)
	DISPATCH ("chain",     handle_chain)
	DISPATCH ("select",    handle_select)
	DISPATCH ("insert",    handle_insert)
	DISPATCH ("where",     handle_where)

	DISPATCH ("refresh",   handle_refresh)
	
	DISPATCH ("pipe",      handle_pipe)
	DISPATCH ("src",       handle_src)
	DISPATCH ("tgt",       handle_tgt)

#ifdef SUPPORT_OBSOLETE_MARKUP
	/* obsolete handlers */
	DISPATCH ("wtokey",    handle_wtokey)
	DISPATCH ("wtoobj",    handle_wtoobj)
#endif /* SUPPORT_OBSOLETE_MARKUP */
	{
		found = FALSE;

		/* Register the plugins as well */
		GList *node;
		for (node=par->element_plugins; node; node=node->next)
		{
			DuiParserElementPlugin *plg = node->data;
  			if (!strcmp (plg->eltname, name)) 
			{
     			if (plg->start_handler)
				{
     				plg->start_handler (plg->user_data, (const char **) attrs); 
				}
				found = TRUE;
				break;
			}
		}
	}
	if (!found)
	{
		int i;
		PERR ("unknown element %s with attrs", name);
		i = 0;
		while (attrs && attrs[i])
		{
			printf ("\tattr=%s\n", attrs[i]);
			i++;
		}
	}
}


static void 
end_elt (void *user_data, const xmlChar *name)
{
	DuiParser *par = (DuiParser *) user_data;
	PINFO ("</%s>", name);

	TERMINATE ("form",       handle_form_end)
	TERMINATE ("report",     handle_report_end)
	TERMINATE ("row",        handle_row_end)
	TERMINATE ("filter",     handle_filter_end)
	TERMINATE ("pipe",       handle_pipe_end)

	/* Register the plugins as well */
	if (par->element_plugins)
	{
		GList *node;
		for (node=par->element_plugins; node; node=node->next)
		{
			DuiParserElementPlugin *plg = node->data;
  			if (!strcmp (plg->eltname, name)) 
			{
     			if (plg->end_handler)
				{
     				plg->end_handler (plg->user_data); 
				}
				break;
			}
		}
	}
}


static void
cdata_cb (void *user_data, const xmlChar *value, int len)
{
#if 0
	DuiParser *par = (DuiParser *) user_data;
	char *p;
	p = g_strndup (value, len);
	printf ("cdata duude len=%d %s\n", len, p);
	g_free (p);
#endif

}


/* =================================================================== */

DuiParser *
dui_parser_new (DuiInterface *dui)
{
	DuiParser *par;

	par = g_new0 (DuiParser, 1);
	
	par->dui = dui;

	/* current state */
	par->element_plugins = NULL;
	par->field_plugins = NULL;
	par->table_plugins = NULL;
	par->struct_plugins = NULL;
	par->database = NULL;
	par->action = NULL;
	par->report = NULL;
	par->filter = NULL;
	par->row_widget = NULL;
	par->row = NULL;
	
	/* initialize the SAX parser */
	bzero (&par->sax, sizeof (xmlSAXHandler));
	par->sax.startDocument = start_doc;
	par->sax.endDocument = end_doc;
	par->sax.fatalError = fatal_err;
	par->sax.error = err;
	par->sax.warning = warn;
	par->sax.startElement = start_elt;
	par->sax.endElement = end_elt;
	par->sax.characters = cdata_cb;

	return par;
}

void 
dui_parser_destroy (DuiParser *par)
{
	/* current state */
	par->dui = NULL;
	par->database = NULL;
	par->action = NULL;
	par->report = NULL;
	par->filter = NULL;
	if (par->row_widget) g_free (par->row_widget);
	par->row_widget = NULL;
	
	/* XXX fixme: must call destroy callpacks on each plugin */
	g_list_free (par->element_plugins);
	g_list_free (par->field_plugins);
	g_list_free (par->table_plugins);
	g_list_free (par->struct_plugins);
	g_free (par);
}

DuiInterface *
dui_parser_get_interface (DuiParser *par)
{
	if (!par) return NULL;
	return par->dui;
}

void 
dui_parser_register_element_plugin (DuiParser *par, DuiParserElementPlugin *plg)
{
	if (!par ||!plg) return;

	par->element_plugins = g_list_append (par->element_plugins, plg);
}

void 
dui_parser_register_field_plugin (DuiParser *par, DuiParserFieldPlugin *plg)
{
	if (!par ||!plg) return;

	par->field_plugins = g_list_append (par->field_plugins, plg);
}

void 
dui_parser_register_table_plugin (DuiParser *par, DuiParserTablePlugin *plg)
{
	if (!par ||!plg) return;

	par->table_plugins = g_list_append (par->table_plugins, plg);
}

void 
dui_parser_register_struct_plugin (DuiParser *par, DuiParserStructPlugin *plg)
{
	if (!par ||!plg) return;

	par->struct_plugins = g_list_append (par->struct_plugins, plg);
}

void 
dui_parse (DuiParser *par, const char * filename)
{
	if (!par || !filename) return;

	if (0 == g_file_test (filename, G_FILE_TEST_EXISTS))
	{
		FATAL ("Can't find DWI interface description file \'%s\'\n", filename);
		return;
	}
	
	xmlSAXUserParseFile (&par->sax, par, filename);
}

/* ========================== END OF FILE ============================ */
